#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "GetTime.h"
#include "Rand.h"
#include "RakPeerInterface.h"
#include "PacketEnumerations.h"
#include "RakNetworkFactory.h"
#include "DataReplicator.h"
#include "NetworkIDGenerator.h"
#include "GetTime.h"
#include <assert.h>
#include <conio.h>

#ifdef _WIN32
#include <windows.h> // Sleep
#else
#include <unistd.h> // usleep
#endif

static const int NUM_PEERS=4;
RakPeerInterface *rakPeer[NUM_PEERS];
DataReplicator messageHandler[NUM_PEERS];

// Implement an interface on some data, so we can control how it is synchronized.  We also could have just passed the count variable directly,
// and the memory synchronizer would revert to doing a simple memcmp and memcpy
class TestClassData : public SynchronizedMemory
{
public:
	TestClassData()
	{
		serializeCB=TestClassData::Serialize;
		deserializeCB=TestClassData::Deserialize;
		copyCB=TestClassData::Copy;
		copyFromCB=TestClassData::CopyFrom;
		freeMemoryCB=TestClassData::FreeMemory;
		shouldSendUpdateCB=TestClassData::ShouldSendUpdate;
		isUnifiedMemoryCB=TestClassData::IsUnifiedMemory;
		includeTimestampCB=TestClassData::IncludeTimestamp;
		autoDeleteOnUnreferenceCB=TestClassData::AutoDeleteOnUnreference;
	}

	// See DataReplicator.h for a detailed explanation of each implementation
	static void Serialize(SynchronizedMemory *memory, void *context, RakNet::BitStream *bitstream, SynchronizedMemory *lastSentValue, unsigned int lastSendTime, PlayerID playerId)
	{
		TestClassData *testClassData = (TestClassData *)memory;
		bitstream->WriteCompressed(testClassData->count);
	}
	static void Deserialize(SynchronizedMemory *memory, void *context, RakNet::BitStream *bitstream, unsigned int timePacketSent, PlayerID playerId)
	{
		TestClassData *testClassData = (TestClassData *)memory;
		bool success;
		success=bitstream->ReadCompressed(testClassData->count);
#ifdef _DEBUG
		assert(success);
#endif
	}
	static SynchronizedMemory * Copy(SynchronizedMemory *memory, void *context)
	{
		TestClassData *testClassData = (TestClassData *)memory;
		TestClassData *tc = new TestClassData;
		// Only need to copy the synchronized data, not everything
		tc->count=testClassData->count;
		return tc;
	}
	static void CopyFrom(SynchronizedMemory *memory, void *context,SynchronizedMemory *source)
	{
		TestClassData *testClassData = (TestClassData *)memory;
		TestClassData *tc = (TestClassData*)source;
		// Only need to copy the synchronized data, not everything
		testClassData->count=tc->count;
	}
	static void FreeMemory(SynchronizedMemory *memory, void *context)
	{
		delete (TestClassData *)memory;
	}
	static bool ShouldSendUpdate(SynchronizedMemory *memory, void *context,PlayerID destinationSystem, SynchronizedMemory *lastSentValue, unsigned int lastSendTime)
	{
		TestClassData *testClassData = (TestClassData *)memory;
		TestClassData *tc = (TestClassData*)lastSentValue;
		return tc->count!=testClassData->count;
	}
	static bool IsUnifiedMemory(SynchronizedMemory *memory, void *context)
	{
		// Since we are sending the data at different times, return false so ShouldSendUpdate gets the correct lastSendValue.
		// If we returned true then all remote systems would share the same pointer
		return false;
	}
	static bool IncludeTimestamp(SynchronizedMemory *memory, void *context)
	{
		// Test sending a timestamp, although I don't use it
		return true;
	}

	static bool AutoDeleteOnUnreference(SynchronizedMemory *memory, void *context)
	{
		return false;
	}

	// The data that is synchronized
	int count;
};

class DataAggregate : public TestClassData
{
public:

	DataAggregate()
	{
		copyCB=DataAggregate::Copy;
		copyFromCB=DataAggregate::CopyFrom;
		serializeCB=DataAggregate::Serialize;
        deserializeCB=DataAggregate::Deserialize;
		shouldSendUpdateCB=DataAggregate::ShouldSendUpdate;
	}

	static SynchronizedMemory * Copy(SynchronizedMemory *memory, void *context)
	{
		DataAggregate *testClassData = (DataAggregate *)memory;
		DataAggregate *tc = new DataAggregate;
		// Only need to copy the synchronized data, not everything
		tc->count=testClassData->count;
		tc->moreData=testClassData->moreData;
		return tc;
	}
	static void CopyFrom(SynchronizedMemory *memory, void *context,SynchronizedMemory *source)
	{
		DataAggregate *testClassData = (DataAggregate *)memory;
		DataAggregate *tc = (DataAggregate*)source;
		// Only need to copy the synchronized data, not everything
		testClassData->count=tc->count;
		testClassData->moreData=tc->moreData;
	}
	static void Serialize(SynchronizedMemory *memory, void *context, RakNet::BitStream *bitstream, SynchronizedMemory *lastSentValue, unsigned int lastSendTime, PlayerID playerId)
	{
		DataAggregate *testClassData = (DataAggregate *)memory;
		// Take advantage of the last send value to send more efficiently.  This is just here as an example.
		DataAggregate *dat = (DataAggregate*)lastSentValue;
		if (dat->count!=testClassData->count)
		{
			bitstream->Write(true);
			bitstream->Write(testClassData->count);
		}
		else
		{
			bitstream->Write(false);
		}

		if (dat->moreData!=testClassData->moreData)
		{
			bitstream->Write(true);
			bitstream->Write(testClassData->moreData);
		}
		else
		{
			bitstream->Write(false);
		}
	}
	static void Deserialize(SynchronizedMemory *memory, void *context, RakNet::BitStream *bitstream, unsigned int timePacketSent, PlayerID playerId)
	{
		DataAggregate *testClassData = (DataAggregate *)memory;
		bool success;
		bool dataWritten;
		bitstream->Read(dataWritten);
		if (dataWritten)
			bitstream->Read(testClassData->count);
		success=bitstream->Read(dataWritten);
		if (dataWritten)
			success=bitstream->Read(testClassData->moreData);

#ifdef _DEBUG
		assert(success);
#endif
	}
	static bool ShouldSendUpdate(SynchronizedMemory *memory, void *context,PlayerID destinationSystem, SynchronizedMemory *lastSentValue, unsigned int lastSendTime)
	{
		DataAggregate *testClassData = (DataAggregate *)memory;
		DataAggregate *dat = (DataAggregate*)lastSentValue;
		return dat->count!=testClassData->count || dat->moreData!=testClassData->moreData;
	}

	char moreData;
};

// A class that holds a synchronized data element.  This is used to track the last value of that data element, so we can print updates when the
// memory synchronizer changes it.
class TestClass 
{
public:
	TestClass()
	{
		countClass.count=lastCount=RakNet::GetTime();
	}

	void SetSender(bool b) 
	{
		sender=b;
	}

	TestClassData *GetSynchronizedDataPtr(void) {return &countClass;}

	void Update(void)
	{
		if (sender)
		{
			countClass.count=RakNet::GetTime();
		}
		else
		{
			if (countClass.count!=lastCount)
			{
				// So we can see when count changes.
				printf("%p: Count changed from %i to %i\n", &(countClass.count), lastCount, countClass.count);
				lastCount=countClass.count;
			}
		}
	}


	// One system updates a count variable, which should be synchronized to other systems at different types.
	// lastCount is used to track when the data changes
	DataAggregate countClass;
	int lastCount;

	// Is this the system updating count?
	bool sender;
};

// All data is global to make testing easier
TestClass testClass[NUM_PEERS];
int i;
unsigned int time;
int peerIndex;

void TestServerTopology(void);
void TestRelayTopology(void);
void Reset(void);
void RunTest(int duration);

int main(void)
{
	for (i=0; i < NUM_PEERS; i++)
		rakPeer[i]=RakNetworkFactory::GetRakPeerInterface();

	printf("Memory Synchronization Demo.\n");

	// Initialize the message handlers
	for (peerIndex=0; peerIndex < NUM_PEERS; peerIndex++)
	{
		rakPeer[peerIndex]->AttachMessageHandler(&(messageHandler[peerIndex]));
		rakPeer[peerIndex]->SetMaximumIncomingConnections(NUM_PEERS);
	}

	TestServerTopology();
	TestRelayTopology();

	printf("Press enter to continue.\n");
	char str[256];
	gets(str);

	for (i=0; i < NUM_PEERS; i++)
		RakNetworkFactory::DestroyRakPeerInterface(rakPeer[i]);

	return 1;
}

void RunTest(int duration)
{
	time=RakNet::GetTime()+duration;

	while (RakNet::GetTime()<time)
	{
		for (i=0; i < NUM_PEERS; i++)
			testClass[i].Update();

		for (i=0; i < NUM_PEERS; i++)
		{
			Packet *p = rakPeer[i]->Receive();
			if (p)
			{
				if (p->data[0]==ID_CONNECTION_REQUEST_ACCEPTED)
					printf("Connection request accepted from %i\n", p->playerId.port);
				rakPeer[i]->DeallocatePacket(p);
			}			
		}

#ifdef WIN32
		Sleep(30);
#else
		usleep(30*1000);
#endif
	}
}

void Reset(void)
{
	for (i=0; i < NUM_PEERS; i++)
		messageHandler[i].Clear();

	for (peerIndex=0; peerIndex < NUM_PEERS; peerIndex++)
	{
		rakPeer[peerIndex]->Disconnect(100);
		messageHandler[peerIndex].RemoveParticipant(UNASSIGNED_PLAYER_ID);
	}

	// Initialize the peers
	for (peerIndex=0; peerIndex < NUM_PEERS; peerIndex++)
	{
		rakPeer[peerIndex]->Initialize(NUM_PEERS, 60000+peerIndex, 0);
	}

	for (i=0; i < NUM_PEERS; i++)
	{
		testClass[i].countClass.count=0;
		testClass[i].lastCount=0;
	}
}

void TestServerTopology(void)
{
	Reset();

	printf("Testing server by having each peer connect to peer 0, and only peer 0\nupdates the variable\n");

	// Connect each peer to peer 0
	for (i=1; i < NUM_PEERS; i++)
	{
		rakPeer[i]->Connect("127.0.0.1", 60000, 0, 0);
	}

	// Wait for the connection
#ifdef WIN32
	Sleep(1000);
#else
	usleep(1000*1000);
#endif

	for (i=0; i < NUM_PEERS; i++)
	{
		// TODO - document this call
		messageHandler[i].SynchronizeMemory("TestClass::count", testClass[i].GetSynchronizedDataPtr(),HIGH_PRIORITY,RELIABLE,0,  i==0, i!=0, i==0 ? 1000 : 0, UNASSIGNED_OBJECT_ID);
	}

	// Let class instance 0 update the data and the others just receive updates
	testClass[0].SetSender(true);
	for (i=1; i < NUM_PEERS; i++)
		testClass[i].SetSender(false);


	RunTest(5000);
}

void TestRelayTopology(void)
{
	Reset();

	printf("Testing relay by having each peer connect to the next higher numbered peer\nand only peer 0 updating the variable\n");

	// Connect each peer to the next peer
	for (i=0; i < NUM_PEERS; i++)
	{
		if (i+1!=NUM_PEERS)
			rakPeer[i]->Connect("127.0.0.1", 60000+i+1, 0, 0);
	}

	// Wait for the connection
#ifdef WIN32
	Sleep(1000);
#else
	usleep(1000*1000);
#endif

	PlayerID playerId;

	for (i=0; i < NUM_PEERS; i++)
	{
		// All systems send and receive.
		messageHandler[i].SynchronizeMemory("TestClass::count", testClass[i].GetSynchronizedDataPtr(),HIGH_PRIORITY,RELIABLE,0,  true, true, i==0 ? 1000 : 0, UNASSIGNED_OBJECT_ID);

		if (i-1>=0)
		{
			rakPeer[i]->IPToPlayerID("127.0.0.1", 60000+i-1, &playerId);
		}

		if (i+1 < NUM_PEERS)
		{
			rakPeer[i]->IPToPlayerID("127.0.0.1", 60000+i+1, &playerId);
		}
	}

	// Let class instance 0 update the data.  The others only track when the data is changed
	testClass[0].SetSender(true);
	for (i=1; i < NUM_PEERS; i++)
		testClass[i].SetSender(false);

	RunTest(10000);
}